﻿'Use VS to create a Visual Basic Console Application named SimulatedBeeColony.
'Replace the contents of file Program.vb with this code.

Imports Microsoft.VisualBasic

Namespace SimulatedBeeColony
    Friend Class Program
        Shared Sub Main(ByVal args() As String)
            Try
                Console.WriteLine(vbLf & "Begin Simulated Bee Colony algorithm demo" & vbLf)
                Console.WriteLine("Loading cities data for SBC Traveling Salesman Problem analysis")
                Dim citiesData As New CitiesData(20)
                Console.WriteLine(citiesData.ToString())
                Console.WriteLine("Number of cities = " & citiesData.cities.Length)
                Console.WriteLine("Number of possible paths = " & citiesData.NumberOfPossiblePaths().ToString("#,###"))
                Console.WriteLine("Best possible solution (shortest path) length = " & citiesData.ShortestPathLength().ToString("F4"))

                Dim totalNumberBees As Integer = 100
                Dim numberInactive As Integer = 20
                Dim numberActive As Integer = 50
                Dim numberScout As Integer = 30

                Dim maxNumberVisits As Integer = 100
                Dim maxNumberCycles As Integer = 3460
                'Dim maxNumberVisits As Integer = 95
                'Dim maxNumberCycles As Integer = 10570
                'Dim maxNumberVisits As Integer = 300 ' proportional to # of possible neighbors to given solution
                'Dim maxNumberCycles As Integer = 32450

                Dim hive As New Hive(totalNumberBees, numberInactive, numberActive, numberScout, maxNumberVisits, maxNumberCycles, citiesData)
                Console.WriteLine(vbLf & "Initial random hive")
                Console.WriteLine(hive)

                Dim doProgressBar As Boolean = True
                hive.Solve(doProgressBar)

                Console.WriteLine(vbLf & "Final hive")
                Console.WriteLine(hive)

                Console.WriteLine("End Simulated Bee Colony demo")
                Console.ReadLine()
            Catch ex As Exception
                Console.WriteLine("Fatal: " & ex.Message)
                Console.ReadLine()
            End Try
        End Sub 'Main()
    End Class 'class Program

    Friend Class Hive
        Public Class Bee
            Public status As Integer '0 = inactive, 1 = active, 2 = scout
            Public memoryMatrix() As Char 'problem-specific. a path of cities.
            Public measureOfQuality As Double 'smaller values are better. total distance of path.
            Public numberOfVisits As Integer

            Public Sub New(ByVal status As Integer,
                           ByVal memoryMatrix() As Char,
                           ByVal measureOfQuality As Double,
                           ByVal numberOfVisits As Integer
                           )
                Me.status = status
                Me.memoryMatrix = New Char(memoryMatrix.Length - 1) {}
                Array.Copy(memoryMatrix, Me.memoryMatrix, memoryMatrix.Length)
                Me.measureOfQuality = measureOfQuality
                Me.numberOfVisits = numberOfVisits
            End Sub

            Public Overrides Function ToString() As String
                Dim s = ""
                s &= "Status = " & Me.status & vbLf
                s &= " Memory = " & vbLf
                For i = 0 To Me.memoryMatrix.Length - 2
                    s &= AscW(Me.memoryMatrix(i)) & "->"
                Next i
                s &= AscW(Me.memoryMatrix(Me.memoryMatrix.Length - 1)) & vbLf
                s &= " Quality = " & Me.measureOfQuality.ToString("F4")
                s &= " Number visits = " & Me.numberOfVisits
                Return s
            End Function
        End Class 'Bee

        Private Shared random As Random = Nothing 'multipurpose

        Public citiesData As CitiesData 'this is the problem-specific data we want to optimize

        Public totalNumberBees As Integer 'mostly for readability in the object constructor call
        Public numberInactive As Integer
        Public numberActive As Integer
        Public numberScout As Integer

        Public maxNumberCycles As Integer 'one cycle represents an action by all bees in the hive
        'public int maxCyclesWithNoImprovement; // deprecated

        Public maxNumberVisits As Integer 'max number of times bee will visit a given food source without finding a better neighbor
        Public probPersuasion As Double = 0.9 'probability inactive bee is persuaded by better waggle solution
        Public probMistake As Double = 0.01 'probability an active bee will reject a better neighbor food source OR accept worse neighbor food source

        Public bees() As Bee
        Public bestMemoryMatrix() As Char 'problem-specific
        Public bestMeasureOfQuality As Double
        Public indexesOfInactiveBees() As Integer ' contains indexes into the bees array

        Public Overrides Function ToString() As String
            Dim s = ""
            s &= "Best path found: "
            For i = 0 To Me.bestMemoryMatrix.Length - 2
                s &= AscW(Me.bestMemoryMatrix(i)) & "->"
            Next i
            s &= AscW(Me.bestMemoryMatrix(Me.bestMemoryMatrix.Length - 1)) & vbLf

            s &= "Path quality:    "
            If bestMeasureOfQuality < 10000.0 Then
                s &= bestMeasureOfQuality.ToString("F4") & vbLf
            Else
                s &= bestMeasureOfQuality.ToString("#.####e+00")
            End If
            s &= vbLf
            Return s
        End Function

        Public Sub New(ByVal totalNumberBees As Integer, ByVal numberInactive As Integer, ByVal numberActive As Integer, ByVal numberScout As Integer, ByVal maxNumberVisits As Integer, ByVal maxNumberCycles As Integer, ByVal citiesData As CitiesData)
            random = New Random(0)

            Me.totalNumberBees = totalNumberBees
            Me.numberInactive = numberInactive
            Me.numberActive = numberActive
            Me.numberScout = numberScout
            Me.maxNumberVisits = maxNumberVisits
            Me.maxNumberCycles = maxNumberCycles
            'Me.maxCyclesWithNoImprovement = maxCyclesWithNoImprovement

            'Me.citiesData = New CitiesData(citiesData.cities.Length) ' hive's copy of problem-specific data
            Me.citiesData = citiesData ' reference to CityData

            'Me.probPersuasion And Me.probMistake are hard-coded in class definition

            Me.bees = New Bee(totalNumberBees - 1) {}
            Me.bestMemoryMatrix = GenerateRandomMemoryMatrix() 'alternative initializations are possible
            Me.bestMeasureOfQuality = MeasureOfQuality(Me.bestMemoryMatrix)

            Me.indexesOfInactiveBees = New Integer(numberInactive - 1) {} 'indexes of bees which are currently inactive

            For i = 0 To totalNumberBees - 1 'initialize each bee, and best solution
                Dim currStatus As Integer 'depends on i. need status before we can initialize Bee
                If i < numberInactive Then
                    currStatus = 0 'inactive
                    indexesOfInactiveBees(i) = i 'curr bee is inactive
                ElseIf i < numberInactive + numberScout Then
                    currStatus = 2 'scout
                Else
                    currStatus = 1 'active
                End If

                Dim randomMemoryMatrix() As Char = GenerateRandomMemoryMatrix()
                Dim mq As Double = MeasureOfQuality(randomMemoryMatrix)
                Dim numberOfVisits = 0

                bees(i) = New Bee(currStatus, randomMemoryMatrix, mq, numberOfVisits) ' instantiate current bee

                'does this bee have best solution?
                If bees(i).measureOfQuality < bestMeasureOfQuality Then ' curr bee is better (< because smaller is better)
                    Array.Copy(bees(i).memoryMatrix, Me.bestMemoryMatrix, bees(i).memoryMatrix.Length)
                    Me.bestMeasureOfQuality = bees(i).measureOfQuality
                End If
            Next i 'each bee

        End Sub 'TravelingSalesmanHive ctor


        Public Function GenerateRandomMemoryMatrix() As Char()
            Dim result(Me.citiesData.cities.Length - 1) As Char ' // problem-specific
            Array.Copy(Me.citiesData.cities, result, Me.citiesData.cities.Length)

            For i = 0 To result.Length - 1 ' Fisher-Yates (Knuth) shuffle
                Dim r As Integer = random.Next(i, result.Length)
                Dim temp As Char = result(r)
                result(r) = result(i)
                result(i) = temp
            Next i
            Return result
        End Function 'GenerateRandomMemoryMatrix()

        Public Function GenerateNeighborMemoryMatrix(ByVal memoryMatrix() As Char) As Char()
            Dim result(memoryMatrix.Length - 1) As Char
            Array.Copy(memoryMatrix, result, memoryMatrix.Length)

            Dim ranIndex = random.Next(0, result.Length) '[0, Length-1] inclusive
            Dim adjIndex As Integer
            If ranIndex = result.Length - 1 Then
                adjIndex = 0
            Else
                adjIndex = ranIndex + 1
            End If

            Dim tmp As Char = result(ranIndex)
            result(ranIndex) = result(adjIndex)
            result(adjIndex) = tmp

            Return result
        End Function 'GenerateNeighborMemoryMatrix()

        Public Function MeasureOfQuality(ByVal memoryMatrix() As Char) As Double
            Dim answer As Double = 0.0
            For i = 0 To memoryMatrix.Length - 2
                Dim c1 = memoryMatrix(i)
                Dim c2 = memoryMatrix(i + 1)
                Dim d As Double = Me.citiesData.Distance(c1, c2)
                answer += d
            Next i
            Return answer
        End Function 'MeasureOfQuality()

        Public Sub Solve(ByVal doProgressBar As Boolean) 'find best Traveling Salesman Problem solution
            Dim pb As Boolean = doProgressBar 'just want a shorter variable
            Dim numberOfSymbolsToPrint As Integer = 10 '10 units so each symbol is 10.0% progress
            Dim increment As Integer = Me.maxNumberCycles \ numberOfSymbolsToPrint
            If pb Then
                Console.WriteLine(vbLf & "Entering SBC Traveling Salesman Problem algorithm main processing loop" & vbLf)
            End If
            If pb Then '10 units so each symbol is 10% progress
                Console.WriteLine("Progress: |==========|")
            End If
            If pb Then
                Console.Write("           ")
            End If
            Dim cycle = 0

            Do While cycle < Me.maxNumberCycles
                For i = 0 To Me.totalNumberBees - 1 'each bee
                    If Me.bees(i).status = 1 Then 'active bee
                        ProcessActiveBee(i)
                    ElseIf Me.bees(i).status = 2 Then 'scout bee
                        ProcessScoutBee(i)
                    ElseIf Me.bees(i).status = 0 Then 'inactive bee
                        ProcessInactiveBee(i)
                    End If
                Next i 'for each bee
                cycle += 1

                'print a progress bar
                If pb AndAlso cycle Mod increment = 0 Then
                    Console.Write("^")
                End If
            Loop 'main while processing loop

            If pb Then 'end the progress bar
                Console.WriteLine("")
            End If
        End Sub 'Solve()

        Private Sub ProcessInactiveBee(ByVal i As Integer)
            Return 'not used in this implementation
        End Sub

        Private Sub ProcessActiveBee(ByVal i As Integer)
            Dim neighbor() = GenerateNeighborMemoryMatrix(bees(i).memoryMatrix) ' find a neighbor solution
            Dim neighborQuality = MeasureOfQuality(neighbor) ' get its quality
            Dim prob = random.NextDouble() 'used to determine if bee makes a mistake; compare against probMistake which has some small value (~0.01)
            Dim memoryWasUpdated = False 'used to determine if bee should perform a waggle dance when done
            Dim numberOfVisitsOverLimit = False 'used to determine if bee will convert to inactive status

            If neighborQuality < bees(i).measureOfQuality Then 'active bee found better neighbor (< because smaller values are better)
                If prob < probMistake Then 'bee makes mistake: rejects a better neighbor food source
                    bees(i).numberOfVisits += 1 'no change to memory but update number of visits
                    If bees(i).numberOfVisits > maxNumberVisits Then
                        numberOfVisitsOverLimit = True
                    End If
                Else 'bee does not make a mistake: accepts a better neighbor
                    Array.Copy(neighbor, bees(i).memoryMatrix, neighbor.Length) 'copy neighbor location into bee's memory
                    bees(i).measureOfQuality = neighborQuality 'update the quality
                    bees(i).numberOfVisits = 0 'reset counter
                    memoryWasUpdated = True 'so that this bee will do a waggle dance
                End If
            Else 'active bee did not find a better neighbor
                'Console.WriteLine("c");
                If prob < probMistake Then 'bee makes mistake: accepts a worse neighbor food source
                    Array.Copy(neighbor, bees(i).memoryMatrix, neighbor.Length) 'copy neighbor location into bee's memory
                    bees(i).measureOfQuality = neighborQuality 'update the quality
                    bees(i).numberOfVisits = 0 'reset
                    memoryWasUpdated = True 'so that this bee will do a waggle dance
                Else 'no mistake: bee rejects worse food source
                    bees(i).numberOfVisits += 1
                    If bees(i).numberOfVisits > maxNumberVisits Then
                        numberOfVisitsOverLimit = True
                    End If
                End If
            End If

            'at this point we need to determine a.) if the number of visits has been exceeded in which case bee becomes inactive
            'or b.) memory was updated in which case check to see if new memory is a global best, and then bee does waggle dance
            'or c.) neither in which case nothing happens (bee just returns to hive).

            If numberOfVisitsOverLimit = True Then
                bees(i).status = 0 'current active bee transitions to inactive
                bees(i).numberOfVisits = 0 'reset visits (and no change to this bees memory)
                Dim x As Integer = random.Next(numberInactive) ' pick a random inactive bee. x is an index into a list, not a bee ID
                bees(indexesOfInactiveBees(x)).status = 1 ' make it active
                indexesOfInactiveBees(x) = i ' record now-inactive bee 'i' in the inactive list
            ElseIf memoryWasUpdated = True Then 'current bee returns and performs waggle dance
                'first, determine if the new memory is a global best. note that if bee has accepted a worse food source this can't be true
                If bees(i).measureOfQuality < Me.bestMeasureOfQuality Then ' the modified bee's memory is a new global best (< because smaller is better)
                    Array.Copy(bees(i).memoryMatrix, Me.bestMemoryMatrix, bees(i).memoryMatrix.Length) ' update global best memory
                    Me.bestMeasureOfQuality = bees(i).measureOfQuality ' update global best quality
                End If
                DoWaggleDance(i)
            Else 'number visits is not over limit and memory was not updated so do nothing (return to hive but do not waggle)
                Return
            End If
        End Sub 'ProcessActiveBee()

        Private Sub ProcessScoutBee(ByVal i As Integer)
            Dim randomFoodSource() As Char = GenerateRandomMemoryMatrix() ' scout bee finds a random food source...
            Dim randomFoodSourceQuality As Double = MeasureOfQuality(randomFoodSource) ' and examines its quality
            If randomFoodSourceQuality < bees(i).measureOfQuality Then ' scout bee has found a better solution than its current one (< because smaller measure is better)
                Array.Copy(randomFoodSource, bees(i).memoryMatrix, randomFoodSource.Length) ' unlike active bees, scout bees do not make mistakes
                bees(i).measureOfQuality = randomFoodSourceQuality
                'no change to scout bee's numberOfVisits or status

                'did this scout bee find a better overall/global solution?
                If bees(i).measureOfQuality < bestMeasureOfQuality Then ' yes, better overall solution (< because smaller is better)
                    Array.Copy(bees(i).memoryMatrix, Me.bestMemoryMatrix, bees(i).memoryMatrix.Length) ' copy scout bee's memory to global best
                    Me.bestMeasureOfQuality = bees(i).measureOfQuality
                End If 'better overall solution

                DoWaggleDance(i) 'scout returns to hive and does waggle dance

            End If 'if scout bee found better solution
        End Sub 'ProcessScoutBee()

        Private Sub DoWaggleDance(ByVal i As Integer)
            For ii = 0 To numberInactive - 1 ' each inactive/watcher bee
                Dim b = indexesOfInactiveBees(ii) ' index of an inactive bee
                If bees(b).status <> 0 Then
                    Throw New Exception("Catastrophic logic error when scout bee waggles dances")
                End If
                If bees(b).numberOfVisits <> 0 Then
                    Throw New Exception("Found an inactive bee with numberOfVisits != 0 in Scout bee waggle dance routine")
                End If
                If bees(i).measureOfQuality < bees(b).measureOfQuality Then ' scout bee has a better solution than current inactive/watcher bee (< because smaller is better)
                    Dim p = random.NextDouble() 'will current inactive bee be persuaded by scout's waggle dance?
                    If Me.probPersuasion > p Then 'this inactive bee is persuaded by the scout (usually because probPersuasion is large, ~0.90)
                        Array.Copy(bees(i).memoryMatrix, bees(b).memoryMatrix, bees(i).memoryMatrix.Length)
                        bees(b).measureOfQuality = bees(i).measureOfQuality
                    End If 'inactive bee has been persuaded
                End If 'scout bee has better solution than watcher/inactive bee
            Next ii ' ach inactive bee
        End Sub 'DoWaggleDance()

    End Class 'class ShortestPathHive

    Friend Class CitiesData
        Public cities() As Char

        Public Sub New(ByVal numberCities As Integer)
            Me.cities = New Char(numberCities - 1) {}
            Me.cities(0) = "A"c
            For i = 1 To Me.cities.Length - 1
                Me.cities(i) = CChar(ChrW(AscW(Me.cities(i - 1)) + 1))
            Next i
        End Sub
        Public Function Distance(ByVal firstCity As Char, ByVal secondCity As Char) As Double
            If firstCity < secondCity Then
                Return 1.0 * (AscW(secondCity) - AscW(firstCity))
            Else
                Return 1.5 * (AscW(firstCity) - AscW(secondCity))
            End If
        End Function
        Public Function ShortestPathLength() As Double
            Return 1.0 * (Me.cities.Length - 1)
        End Function
        Public Function NumberOfPossiblePaths() As Long
            Dim n = Me.cities.Length
            Dim answer = 1L
            For i = 1 To n
                'checked
                answer *= i
                Next i
            Return answer
        End Function
        Public Overrides Function ToString() As String
            Dim s = ""
            s &= "Cities: "
            For i = 0 To Me.cities.Length - 1
                s &= AscW(Me.cities(i)) & " "
            Next i
            Return s
        End Function
    End Class 'class CitiesData

End Namespace 'ns